/* memory.c */

/* ce fichier contient les fonctions du gestionnaire de memoire qui sont
   utilisables par le runtime.
*/

#include "mlvalues.h"
#include "fail.h"
#include "gc.h"
#include "gc_aux.h"
#include "mem_misc.h"
#include "free_list.h"
#include "memory.h"
#include "roots.h"
#include "stat_alloc.h"
#include "alloc.h"

#ifdef ANSI
static address grow_heap (unsigned long);
#else
static address grow_heap ();
#endif

address heap_end, heap_start;

/* variables globales de la free-list */
extern address next_block, free_list_mark;
extern address *last_link;

/* variables privees du gestionnaire de memoire */
static unsigned long heap_size, copy_size;
char *temp_alloc_ptr;

/* variables de gestion de la generation */
address yp, young_end, young_start;
extern value **ref_index, *ref_table [];
extern int minor_gc_is_running;
extern int out_of_memory;

int gc_is_running;

/* Pour declencher une exception */
extern struct trap_frame * tp;

/***********************************************************************/
/** appel du gc mineur                                                **/
/***********************************************************************/

value *call_minor_gc (request)
     blocksize request;
{
  gc_message_str ("<");
  gc_is_running = 1;
  Zassert (request <= MAX_YOUNG_SIZE);
  minor_gc ();
  ref_index = ref_table;
  yp = (value *) young_start + request;
  gc_message_str (">");
  gc_is_running = 0;
  return young_start;
}

/***********************************************************************/
/** determination des parametres du gc majeur et appel de celui-ci    **/
/***********************************************************************/

extern address to_end, to_start;
extern address from_end, from_start;
extern address copy_ptr;
extern int major_gc_did_occur;

static void call_major_gc (request)
     unsigned long request;
{
  int ok;
  int no_space_left;

  gc_is_running = 1;
  major_gc_did_occur = 1;
  gc_message_str ("GC...");

  /* Quand on est a court de memoire, on coupe la pile ML et on revient ici. */
  major_gc_again :

  /* trouver le from-space */

  from_start = to_end;
  from_end = from_start + (to_end - to_start);
  do{
    ok = true;
    if (from_end > heap_end){	/* on est a la fin du tas */
      from_start = heap_start;
      from_end = from_start + (to_end - to_start);
      if (from_end > to_start){
	from_end = to_start;
      }
    }
    if (SIZE (OBJECT (from_start)) + from_start > from_end){
      /* l'objet est trop gros : on le saute */
      from_start += SIZE (OBJECT (from_start))+ 1;
      from_end = from_start + (to_end - to_start);
      ok = false;
    }
    if (from_start <= to_start && from_end > to_start){
      from_end = to_start;
    }
    if (from_start == to_start){           /* on a fait tout le tour */
      from_start = to_end;
      if (from_start == heap_end){
	from_start = heap_start;
      }
      from_end = from_start + (to_end - to_start);
      if (from_end > heap_end){
	from_end = heap_end;
      }
      ok = true;
    }
  }while (!ok);
  
  /* from-space ne peut pas etre vide. */
  Zassert (from_start < from_end);

#ifdef DEBUG
  /* heap_check ne peut pas etre appellee de n'importe quel endroit. */
   heap_check (1);
#endif

  /*** appel du gc ***/
  major_gc ();

  /* decider ensuite quel sera le prochain to-space */
  /* taille optimale (?) du to-space : 3/4 (?) de l'espace libre. */
  /* (formule a revoir) */
  {
    address new_to_start, new_to_end;
    unsigned long opti_size = 
      3 * heap_size * (copy_size - (copy_ptr - to_start)) / copy_size / 4;

    if (opti_size < heap_size / 50){
      opti_size = heap_size / 50;
    }
    if (opti_size < 1024){
      opti_size = 1024;
    }
    if (opti_size > heap_size / 2){
      opti_size = heap_size / 2;
    }
    if (from_end == heap_end){
      new_to_end = heap_end;
      new_to_start = heap_end - opti_size;
    }else{
      new_to_start = from_start;
      new_to_end = new_to_start + opti_size;
    }
    if (new_to_end > from_end){
      new_to_end = from_end;
    }
    if (new_to_start < from_start){
      new_to_start = from_start;
    }

    if (new_to_start > from_start){
      ADD_TO_FREE_LIST (from_start, new_to_start - from_start);
    }
    if (new_to_end < from_end){
      ADD_TO_FREE_LIST (new_to_end, from_end - new_to_end);
    }
    
    to_end = new_to_end;
    to_start = new_to_start;

    copy_size = to_end - to_start;
  }

  /* Agrandir le tas si le taux d'occupation est trop grand ou si on
     n'a pas un bloc assez grand pour satisfaire la requete.
  */

  no_space_left = 0;
  if (get_biggest_free_block_size () < request /* || ... */ ){
    address adr;
    unsigned long growing = (request + HEAP_GROWING) / HEAP_GROWING;

    adr = grow_heap (growing * HEAP_GROWING);
    if (adr == (address) -1){
      no_space_left = 1;
    }else{
      ADD_TO_FREE_LIST (adr, growing * HEAP_GROWING);
      gc_message ("Allocating ",
                  growing * HEAP_GROWING * sizeof (value) / 1024,
                  "k for heap zone.");
    }
  }

  /* le gc laisse la free-list ouverte. A nous de la fermer */
  close_free_list ();

#ifdef DEBUG
  /* heap_check ne peut pas etre appellee de n'importe quel endroit */
  heap_check (0);
#endif

  if (no_space_left && (get_biggest_free_block_size () < request)){
    /* il faut lever l'exception "out of memory" */
    /* il faut appeler extract_next_block pour que la free-list soit dans
       un etat coherent.
    */
    extract_next_block ();
    if (minor_gc_is_running){
      /* Il ne faut pas faire un raise sous les pieds du GC mineur.
	 On coupe donc la pile a la main, et on recommence un GC majeur dans
	 l'espoir d'avoir tue quelques objets. Si ce c'est pas le cas, on
	 recoupera la pile...
	 Il faut faire attention de s'arreter si on est au niveau
	 du toplevel-handler : on ne peut pas faire raise au dessus.
      */
      if (out_of_memory) {
        /* Ca n'est pas la premiere fois qu'on coupe la pile.
           On reessaye avec la prochaine exception. */
        tp = tp->tp;
      }
      if (tp >= (struct trap_frame *) ret_stack_high){
	/* il n'y a plus de handler d'exception dans la chaine */
	fatal_error ("Really out of memory. Cannot complete GC. Sorry.");
      }
      extern_asp = tp->asp;
      extern_rsp = (value *) tp;
      out_of_memory = 1;
      goto major_gc_again;
    }else{
      /* le tas est dans un etat coherent. on peut appeler raise. */
      gc_message_str ("...out of memory.\n");
      gc_is_running = 0;
      system_failwith_out_of_memory ();
    }
  }

  /* to-space ne doit pas etre vide. */
  Zassert (to_end > to_start);

  gc_message_str ("...done.\n");
  gc_is_running = minor_gc_is_running;

}

#ifdef UNIX_MEMORY

static char *current_break;

/* shift_heap decale le tas vers les adresses hautes pour permettre d'agrandir
   la zone d'allocation statique. Deux methodes possibles :
   -- declencher un gc majeur avec les bonnes zones "from" et "to".
   -- copier le tas par bcopy et ajuster tous les pointeurs.
*/

static unsigned long heap_shift_request;

void shift_root (x)
     address x;
{
  Zassert (IS_IN_HEAP (*x) && IS_BLOCK (*x));
    * (char **) x += heap_shift_request;
}

void shift_heap (request, adr, len)
     unsigned long request;
     VOID **adr;
     unsigned long *len;
{
  int err_code;
  unsigned long heap_length;
  blocksize i;

  /* nettoyage des tables de references inverses et autres... */
  (void) call_minor_gc (0);

  if (brk (current_break + request)){
    *adr = NULL;
    *len = 0;
    return;
  }else{
    current_break += request;
    temp_alloc_ptr = current_break;
  }

  heap_shift_request = request;
  
  Zassert (request % sizeof (value) == 0);

				/* Voyons la methode "bcopy". */
  /* On veut le tas dans un etat sympathique. */
  if (next_block != NULL){
    put_back_current_block ();
  }else if (block_end > hp){
    HEADER (OBJECT (hp)) =
                 MAKE_HEADER (block_end - hp - 1, TAG_JUNK);
  }

  gc_message (" Shifting heap by ", request / 1024, "k ...");

  heap_length = (char *) heap_end - (char *) heap_start;
  bcopy ((char *) heap_start, (char *) heap_start + request, heap_length);

  { address cur;		/* Ajustement des pointeurs. */
    for (cur = (address) ((char *) heap_start + request);
	 cur < (address) ((char *) heap_end + request); ){
      if (cur == (address) ((char *) to_start + request)){
	cur = (address) ((char *) to_end + request);
      } else {
	register unsigned long len = H_SIZE (cur);
        if (H_TAG (cur) < NO_SCAN_TAG){
          cur += 1;
          for (i = 0; i < len; i++, cur++)
            if (IS_IN_HEAP (*cur) && IS_BLOCK (*cur))
              *((char **) cur) += request;
        } else {
          cur += len + 1;
        }
      }
    }
    ALL_ROOTS (shift_root);	/* Idem pour les racines. */
  }

  *adr = (VOID *) heap_start;
  *len = request;

  heap_start = (address) ((char *) heap_start + request);
  heap_end = (address) ((char *) heap_end + request);
  to_start = (address) ((char *) to_start + request);
  to_end = (address) ((char *) to_end + request);
  temp_free ();
				/* variables de la free-list */
  free_list_mark = (address) ((char *) free_list_mark + request);
  hp = (address) ((char *) hp + request);
  block_end = (address) ((char *) block_end + request);
  if (next_block != NULL){
    address *cur;
				/* ajuster les chainages de la free-list */
    next_block = (address) ((char *) next_block + request);
    cur = (address *) next_block;
    do{
      LINK (cur) = (address) ((char *) LINK (cur) + request);
      cur = (address *) LINK (cur);
    }while (cur != (address *) next_block);
    last_link = (address *) ((char *) last_link + request);
    extract_next_block ();
  }

  gc_message_str (" done.\n");
}

#endif /* UNIX_MEMORY */

/***************************************************************************/
/*  fonctions d'allocation                                                 */
/***************************************************************************/


/* allocation en first-fit circulaire. Ce code est appele par le gc mineur */

value alloc_in_free_list (len)  /* longueur TOTALE (avec header.) */
     blocksize len;
{
  if (next_block != NULL){
    put_back_current_block ();
  search:
    free_list_mark = next_block;
    do{
      if (SIZE (OBJECT (next_block)) + 1 >= len){
	extract_next_block ();
	hp += len;
	return OBJECT (hp - len);
      }
      last_link = &LINK (next_block);
      next_block = *last_link;
    }while (next_block != free_list_mark && next_block != NULL);
  }else if (block_end > hp){
    HEADER (OBJECT (hp)) =
                 MAKE_HEADER (block_end - hp - 1, TAG_JUNK);
  }
  call_major_gc (len);
  goto search;
}

/* pour allouer, on a le choix entre :
   la macro ALLOC pour des blocs plus petits que MAX_YOUNG_SIZE
     (elle est definie dans alloc.h)
   la fonction big_alloc pour allouer les chaines dans la generation vieille
   la fonction big_alloc_init pour allouer les vecteurs dans la generation
      vieille en les initialisant.
   la fonction temp_alloc, reservee aux fonctions C, qui donne de la memoire
      temporaire, susceptible d'etre ecrasee sans preavis au prochain GC.
   la fonction stat_alloc, qui alloue en zone statique (non collectee)
      Elle est definie dans stat_alloc.c
*/

value big_alloc (hdr, len)
     header hdr;
     blocksize len;
{
  value result = OBJECT (hp);

  Zassert (len <= MAX_SIZE);
  hp += len + 1;
  if (hp > block_end){
    hp -= len + 1;
    result = alloc_in_free_list (len + 1);
  }
  HEADER (result) = hdr;
  Zassert (SIZE (result) == len);
  return result;
}

value big_alloc_init (hdr, len, init)
     header hdr;
     blocksize len;
     value init;
{
  value result;
  PUSH_ROOTS (i, 1);
  i [0] = init;

  Zassert (TAG_HDR (hdr) < NO_SCAN_TAG);
  if (IS_BLOCK (i [0])){
    if (IS_YOUNG (ADDRESS (i [0]))){
      (void) call_minor_gc ((unsigned long) 0);
    }
    Zassert (!IS_YOUNG (ADDRESS (i [0])));
  }
  result = big_alloc (hdr, len);
  while (len--){
    FIELD (result, len) = i [0];
  }
  POP_ROOTS ();
  return result;
}

value big_alloc_init_up (hdr, len)
     header hdr;
     blocksize len;
{
  value * to;
  value init;
  PUSH_ROOTS (result, 1);

  Zassert (TAG_HDR (hdr) < NO_SCAN_TAG);
  result [0] = big_alloc(hdr, len);
  for (to = &FIELD(result [0], 0); len > 0; to++, len--) {
    init = *extern_asp++;
    *to = init;
    if (IS_BLOCK(init) && IS_YOUNG(init)) {
      *ref_index++ = to;
      if (ref_index == &(ref_table[REF_TABLE_SIZE])) {
        (void) call_minor_gc(0);
      }
    }
  }
  POP_ROOTS ();
  return result [0];
}

value big_alloc_init_down (hdr, len)
     header hdr;
     blocksize len;
{
  value * to;
  value init;
  PUSH_ROOTS (result, 1);

  Zassert (TAG_HDR (hdr) < NO_SCAN_TAG);
  result [0] = big_alloc(hdr, len);
  for (to = &FIELD(result [0], len - 1); len > 0; to--, len--) {
    init = *extern_asp++;
    *to = init;
    if (IS_BLOCK(init) && IS_YOUNG(init)) {
      *ref_index++ = to;
      if (ref_index == &(ref_table[REF_TABLE_SIZE])) {
        (void) call_minor_gc(0);
      }
    }
  }
  POP_ROOTS ();
  return result [0];
}

/*******************************************************************
   allocation de memoire temporaire a l'usage des fonctions C
   il est garanti que deux temp_alloc successifs renvoient des portions
   de memoire contigues et que le tout disparait a chaque gc.
********************************************************************/

typedef unsigned long word;

word * static_start, * static_end;

#define MAX(x,y) ((x) < (y) ? (y) : (x))

#ifdef UNIX_MEMORY

char *temp_alloc (size)		/* size en octets */
     blocksize size;
{
  char *result = temp_alloc_ptr;

  if (temp_alloc_ptr + size <= current_break){
    temp_alloc_ptr += size;
    return result;
  }else{
    int err_code;
    long to_allocate = MAX (TEMP_ALLOC_INCR, size);

    err_code = brk (current_break + to_allocate);
    if (!err_code){
      current_break += to_allocate;
      temp_alloc_ptr += size;
      return result;
    }else{
      return 0;
    }
  }
}

void temp_free ()		/* libere tous les blocs temporaires */
{
  temp_alloc_ptr = (char *) heap_end;
}

#else /* not UNIX_MEMORY */

char *temp_alloc (size)		/* size en octets */
     blocksize size;
{
  char *result = temp_alloc_ptr;

  if (temp_alloc_ptr + size <= (char *) heap_start){
    temp_alloc_ptr += size;
    return result;
  }else{
    return 0;
  }
}

void temp_free ()		/* libere tous les blocs temporaires */
{
  temp_alloc_ptr = (char *) static_end;
}

#endif /* UNIX_MEMORY */

/*******************************************************************/
/** agrandir le tas                                               **/
/*******************************************************************/

#ifdef UNIX_MEMORY

static address grow_heap (size)
     unsigned long size;
{
  int err_code;
  address result = heap_end;

  if (current_break < (char *) (heap_end + size)){
    err_code = brk (heap_end + size);
    if (!err_code){
      heap_end += size;
      current_break = (char *) heap_end;
      temp_alloc_ptr = current_break;
      return result;
    }else{
      return (address) -1;
    }
  }else{
    heap_end += size;
    temp_alloc_ptr = (char *) heap_end;
    return result;
  }
}

#else /* not UNIX_MEMORY */

static address grow_heap (size)
     unsigned long size;
{
  if (heap_start - size >= (address) static_end) {
    heap_start -= size;
    return heap_start;
  } else {
    return (address) -1;
  }
}

#endif /* UNIX_MEMORY */

/*******************************************************************/
/** initialisation de la memoire                                  **/
/*******************************************************************/

void init_memory (total_sz, heap_sz, generation_sz, static_sz) 
     unsigned long total_sz;      /* taille de tout l'espace memoire en mots */
     unsigned long heap_sz;       /* taille du tas, en mots */
     unsigned long generation_sz; /* taille de la zone mineure, en mots */
     unsigned long static_sz;     /* taille de la zone statique, en mots */
{
#ifndef UNIX_MEMORY
  char * memory_low, * memory_high;
  memory_low = (char *) (((malloc(total_sz * sizeof(word) + 3) + 3) / 4) * 4);
  if (memory_low == 0)
    fatal_error ("Not enough memory for startup.");
  memory_high = memory_low + total_sz * sizeof(word);
  if (memory_low + static_sz * sizeof(word) + heap_sz * sizeof(header) >
      memory_high)
    fatal_error ("Not enough memory for startup.");
#endif

  /* allocation de la zone statique */

#ifdef UNIX_MEMORY
  static_start = (word *) sbrk (static_sz * sizeof (word));
  static_end = static_start + static_sz;
  if (static_start == (word *) -1){
    fatal_error ("Not enough memory for startup.");
  }
#else
  static_start = (word *) memory_low;
  static_end = static_start + static_sz;
#endif

#ifdef DEBUG
  { unsigned long i;
    for (i = 0; i < static_sz; i++){
      static_start [i] = (word) gc_debug_random ();
    }
  }
#endif

  init_stat_memory (static_start, static_sz * sizeof (word));

  /* allocation de la zone majeure */

#ifdef UNIX_MEMORY
  heap_start = (address) sbrk (heap_sz * sizeof (header));
  heap_end = (address) sbrk (0);
  if (heap_start == (address) -1){
    fatal_error ("Not enough memory for startup.");
  }
#else
  heap_start = (address) memory_high - heap_sz;
  heap_end = (address) memory_high;
#endif

#ifdef DEBUG
  { unsigned long i;
    for (i = 0; i < heap_sz; i++){
      heap_start [i] = (value) gc_debug_random ();
    }
  }
#endif

  /* initialisation des variables du GC majeur */
  heap_size = heap_sz;
  to_start = heap_start;
  to_end = to_start + heap_sz / 2;
  copy_size = heap_sz / 2;
  new_free_list ();
  add_to_free_list (to_end, heap_end - to_end);
  close_free_list ();
  extract_next_block ();
#ifdef DEBUG
  gc_message_str ("\n");
#endif

  /* Allocation de la zone mineure. Elle se fait ici, car elle peut declencher
     un gc majeur par le biais de stat_alloc.
  */
  young_start = (address) stat_alloc (generation_sz * sizeof (value));
  young_end = (address) ((value *) young_start + generation_sz);
  yp = young_start;
  init_minor_gc ();

  /* initialisation des variables du GC mineur */
  ref_index = ref_table;

  /* initialisation des variables de l'allocation temporaire */

#ifdef UNIX_MEMORY
  current_break = (char *) heap_end;
  temp_alloc_ptr = (char *) heap_end;
#else
  temp_alloc_ptr = (char *) static_end;
#endif

  /* Initialisation de la liste des racines temporaires a l'usage des
   * primitives C.
   */
  C_roots_head = NULL;
}
