#open "unix";;
#open "obj";;
#open "misc";;
#open "constants";;
#open "lex_run";;
#open "lex_auto";;
#open "parse_engine";;
#open "parse_tables";;
#open "locations";;
#open "syntax";;
#open "builtins";;
#open "hashtable";;
#open "globals";;
#open "modules";;
#open "types";;
#open "type_errors";;
#open "typing";;
#open "typing_decl";;
#open "typing_intf";;
#open "front";;
#open "instruct";;
#open "back";;

(* Fonctions de parsing *)

let parse_phrase initial_token =
  try
    let res = yy_parse zinc_psyntax Main initial_token in
      clear_v_stack();                  (* Soyons sympa avec le GC *)
      res
  with Parsing curr_tok ->
    let pos1 = start_token()
	and pos2 = end_token() in
    let rec skip tok =
      if (tok != token_end_of_file) & (tok != token_end_of_phrase)
      then skip (Main ())
    in
      skip curr_tok;
      prerr_location (Loc(pos1, pos2));
      prerr_begline ">> Syntax error.";
      prerr_endline "";
      raise Toplevel
;;

let parse_impl_phrase () =
  (magic_obj (parse_phrase (-1)) : impl_phrase)
and parse_intf_phrase () =
  (magic_obj (parse_phrase start_token_Interface) : intf_phrase)
;;

(* Execution des directives *)

let do_directive = function
    Zdir("open", name) ->
      used_modules := find_module name :: !used_modules; ()
  | Zdir("close", name) ->
      used_modules := exceptq (find_module name) !used_modules; ()
  | Zdir("infix", name) ->
      lex_aux2__add_infix name; ()
  | Zdir("uninfix", name) ->
      lex_aux2__remove_infix name; ()
  | Zdir(d, name) ->
      prerr_begline "> Unknown directive \"";
	  prerr_string d;
      prerr_endline2 "\", ignored."
;;

(* Compilation d'une interface *)

let compile_intf_phrase (Intf(desc,loc)) =
  begin match desc with
    Zvaluedecl decl ->
      type_valuedecl loc decl
  | Ztypedecl decl ->
      type_typedecl loc decl
  | Zexcdecl decl ->
      type_excdecl loc decl
  | Zintfdirective dir ->
      do_directive dir
  end;
  flush std_out
;;

let compile_intf name filename =
  let ic = open_in filename in
    new_globalenv name;
    type_stamp := 268435456;
(* Hack pour eviter de donner le meme stamp a un type defini ou declare dans
   l'interface et a un type defini dans l'implementation *)
    init_lexer ic;
    input_name := filename;
    try
      while true do
        compile_intf_phrase(parse_intf_phrase())
      done
    with End_of_file ->
      close_in ic;
      write_compiled_interface()
    | x ->
      close_in ic;
      raise x
;;


(* Compilation d'une implementation *)

let compile_impl_phrase extstream (Impl(desc,loc)) =
  begin match desc with
    Zexpr expr ->
      type_expression loc expr;
      output_value extstream (compile_lambda (translate_expression expr))
  | Zletdef(rec_flag, pat_expr_list) ->
      type_letdef loc rec_flag pat_expr_list;
      output_value extstream
         (if rec_flag
          then compile_lambda(translate_letdef_rec loc pat_expr_list)
          else compile_lambda(translate_letdef loc pat_expr_list))
  | Ztypedef decl ->
      type_typedecl loc decl
  | Zexcdef decl ->
      type_excdecl loc decl
  | Zimpldirective dir ->
      do_directive dir
  end;
  flush std_out
;;

let compile_impl name filename =
  let ic = open_in filename
  and extstream = open_out (name ^ ".zo") in
    init_lexer ic;
    input_name := filename;
    try
      while true do
        compile_impl_phrase extstream (parse_impl_phrase())
      done
    with End_of_file ->
      close_out extstream;
      close_in ic
    | x ->
      close_out extstream;
      unlink (name ^ ".zo");
      close_in ic;
      raise x
;;

let compile_implementation name filename =
  new_globalenv name;
  if file_exists (name ^ ".mli") then
    begin
      let intf = load_module name in
        enter_interface_definitions intf;
        compile_impl name filename;
        check_interface intf
    end
  else
    begin
      compile_impl name filename;
      write_compiled_interface()
    end
;;
